import time
import random

import pytest

from liaa.routing import KBucket, TableTraverser, RoutingTable
from liaa.network import KademliaProtocol

class TestKBucket:
	# pylint: disable=no-self-use
	def test_can_create_bucket(self):
		bucket = KBucket(0, 10, 5)
		assert isinstance(bucket, KBucket)
		assert bucket.last_updated

	def test_can_add_nodes_to_bucket(self, mknode):
		bucket = KBucket(0, 10, 2)
		assert bucket.add_node(mknode()) is True
		assert bucket.add_node(mknode()) is True
		assert bucket.add_node(mknode()) is False
		assert len(bucket) == 2

	def test_get_nodes_returns_proper_nodes(self, mknode):
		bucket = KBucket(0, 10, 2)
		bucket.add_node(mknode(intid=1))
		bucket.add_node(mknode(intid=2))
		fetched = bucket.get_nodes()
		assert len(fetched) == 2

	def test_nodes_past_k_are_pushed_to_replacement(self, mknode):
		k = 3
		bucket = KBucket(0, 10, 3)
		nodes = [mknode() for _ in range(10)]
		for node in nodes:
			bucket.add_node(node)

		# any number of nodes that exceeds `k` should be found in
		# replacement nodes
		replacement_nodes = bucket.replacement_nodes
		assert list(bucket.nodes.values()) == nodes[:k]
		assert list(replacement_nodes.values()) == nodes[k:]

	def test_remove_node_does_nothing_when_node_is_not_in_bucket(self, mknode):
		k = 3
		bucket = KBucket(0, 10, k)
		nodes = [mknode() for _ in range(10)]
		for node in nodes:
			bucket.add_node(node)

		# we remove a node that's not in the bucket (a replacement node) so
		# nothing should change
		replacement_nodes = bucket.replacement_nodes
		bucket.remove_node(nodes.pop())
		assert list(bucket.nodes.values()) == nodes[:k]
		assert list(replacement_nodes.values()) == nodes[k:]

	def test_remove_node_replaces_removed_node_with_replacement_node(self, mknode):
		k = 3
		bucket = KBucket(0, 10, k)
		nodes = [mknode() for _ in range(10)]
		for node in nodes:
			bucket.add_node(node)

		# here we remove a node that's in the bucket, and assert that a
		# our latest replacement node (nodes[-1:]) was added to the bucket
		replacement_nodes = bucket.replacement_nodes
		bucket.remove_node(nodes.pop(0))
		assert list(bucket.nodes.values()) == nodes[:k-1] + nodes[-1:]
		assert list(replacement_nodes.values()) == nodes[k-1:-1]

	def test_remove_all_nodes_uninitializes_bucket(self, mknode):
		k = 3
		bucket = KBucket(0, 10, k)
		nodes = [mknode() for _ in range(10)]
		for node in nodes:
			bucket.add_node(node)

		replacement_nodes = bucket.replacement_nodes

		# remove all nodes
		random.shuffle(nodes)
		for node in nodes:
			bucket.remove_node(node)
		assert not bucket
		assert not replacement_nodes

	def test_split_bucket_regroups_nodes_appropriately(self, mknode):
		bucket = KBucket(0, 10, 5)
		bucket.add_node(mknode(intid=5))
		bucket.add_node(mknode(intid=6))

		one, two = bucket.split()
		assert len(one) == 1
		assert one.range == (0, 5)
		assert len(two) == 1
		assert two.range == (6, 10)

	def test_double_node_is_put_at_end_when_added_twice(self, mknode):
		# make sure when a node is double added it's put at the end
		bucket = KBucket(0, 10, 3)
		nodes = [mknode(), mknode(), mknode()]
		for node in nodes:
			bucket.add_node(node)
		for index, node in enumerate(bucket.get_nodes()):
			assert node == nodes[index]

	def test_has_in_range_works_ok(self, mknode):
		bucket = KBucket(0, 10, 10)
		assert bucket.has_in_range(mknode(intid=5)) is True
		assert bucket.has_in_range(mknode(intid=11)) is False
		assert bucket.has_in_range(mknode(intid=10)) is True
		assert bucket.has_in_range(mknode(intid=0)) is True


class TestRoutingTable:

	# pylint: disable=no-self-use
	def test_can_instantiate_and_flush_table(self, mknode):
		ksize = 3
		table = RoutingTable(KademliaProtocol, ksize=ksize, node=mknode())
		assert isinstance(table, RoutingTable)
		assert len(table.buckets) == 1

	def test_can_split_bucket(self, mknode, mkbucket):
		ksize = 3
		table = RoutingTable(KademliaProtocol, ksize=ksize, node=mknode())
		table.buckets.extend([mkbucket(ksize), mkbucket(ksize)])
		assert len(table.buckets) == 3
		table.split_bucket(0)
		assert len(table.buckets) == 4

	def test_lonely_buckets_returns_stale_buckets(self, mknode, mkbucket):
		ksize = 3
		table = RoutingTable(KademliaProtocol, ksize, node=mknode())
		table.buckets.append(mkbucket(ksize))
		table.buckets.append(mkbucket(ksize))

		# make bucket lonely
		table.buckets[0].last_updated = time.monotonic() - 3600
		lonelies = table.lonely_buckets()
		assert len(lonelies) == 1

	def test_remove_contact_removes_buckets_node(self, mknode, mkbucket):
		ksize = 3
		table = RoutingTable(KademliaProtocol, ksize, node=mknode())
		table.buckets.append(mkbucket(ksize))
		assert len(table.buckets[1]) == 0

		node = mknode()
		table.add_contact(node)
		index = table.get_bucket_index_for(node)
		assert len(table.buckets[index]) == 1

		table.remove_contact(node)
		index = table.get_bucket_index_for(node)
		assert len(table.buckets[index]) == 0

	def test_is_new_node_returns_true_when_node_is_new(self, mknode):
		table = RoutingTable(KademliaProtocol, 3, node=mknode())
		assert table.is_new_node(mknode())

	def test_add_contact_is_ok(self, mknode):
		ksize = 3
		table = RoutingTable(KademliaProtocol, ksize, node=mknode())
		table.add_contact(mknode())
		assert len(table.buckets) == 1
		assert len(table.buckets[0].nodes) == 1

	@pytest.mark.skip(reason="TODO: implement after crawler tests")
	def test_find_neighbors_returns_k_neighbors(self, mknode, _):
		ksize = 3
		_ = RoutingTable(KademliaProtocol, ksize, node=mknode())


# pylint: disable=too-few-public-methods
class TestTableTraverser:
	# pylint: disable=no-self-use
	def test_iteration(self, fake_server, mknode):
		"""
		Make 10 nodes, 5 buckets, two nodes add to one bucket in order

		Bucket 0
			[node0, node1]
		Bucket 1
			[node2, node3]
		Bucket 2
			[node4, node5]
		Bucket 3
			[node6, node7]
		Bucket 4
			[node8, node9]
		Test traver result starting from node4.
		"""

		nodes = [mknode(intid=x) for x in range(10)]

		buckets = []
		for i in range(5):
			bucket = KBucket(2 * i, 2 * i + 1, 2)
			bucket.add_node(nodes[2 * i])
			bucket.add_node(nodes[2 * i + 1])
			buckets.append(bucket)

		# replace router's bucket with our test buckets
		fake_server.router.buckets = buckets

		# expected nodes order
		# pylint: disable=bad-continuation
		expected_nodes = [
			nodes[5],
			nodes[4],
			nodes[3],
			nodes[2],
			nodes[7],
			nodes[6],
			nodes[1],
			nodes[0],
			nodes[9],
			nodes[8],
		]

		start_node = nodes[4]
		table_traverser = TableTraverser(fake_server.router, start_node)
		for index, node in enumerate(table_traverser):
			assert node == expected_nodes[index]



from operator import itemgetter
import heapq
import logging
from typing import Optional, List, Any

from liaa.utils import hex_to_int, check_dht_value_type, digest_to_int


log = logging.getLogger(__name__)  # pylint: disable=invalid-name


# pylint: disable=too-few-public-methods
class NodeType:
	Peer = "peer"
	Resource = "resource"


# pylint: disable=too-many-instance-attributes
class Node:
	# pylint: disable=bad-continuation
	def __init__(self,
		digest_id: bytes,
		ip: Optional[str] = None,
		port: Optional[int] = None,
		type: int = NodeType.Peer,  # pylint: disable=redefined-builtin
		value: Optional[Any] = None
	):
		"""
		Node

		Simple object to encapsulate the concept of a Node (minimally an ID, but
		also possibly an IP and port if this represents a node on the network).
		This class should generally not be instantiated directly, as it is a low
		level construct mostly used by the router.

		A node can either be a peer, or a resource in the network

		Parameters
		----------
			digest_id: bytes
				A value between 0 and 2^160 (as byte array)
			ip: str
				Optional IP address where this Node lives
			port:
				Optional port for this Node (set when IP is set)
			type: int
				Indicator of whether the node represents a peer or a resource in
				the network
			value: Optional[Any]
				Payload associated with node (if self.type == NodeType.Resource)
		"""

		# byte array composed of int_to_digest(self.int_id)
		self.digest_id = digest_id
		self.int_id = digest_to_int(self.digest_id)
		self.hex = self.digest_id.hex()
		self.long_id = hex_to_int(self.digest_id.hex())

		self.ip = ip  # pylint: disable=invalid-name
		self.port = port

		self.type = type
		self.value = value

	def has_valid_value(self) -> bool:
		return check_dht_value_type(self.value)

	def is_same_node(self, node: "Node") -> bool:
		return self.ip == node.ip and self.port == node.port

	def distance_to(self, node: "Node") -> int:
		"""
		Get the distance between this node and another.

		Parameters
		----------
			node: Node
				Node against which to measure key distance
		"""
		return self.long_id ^ node.long_id

	def __eq__(self, other: "Node") -> bool:
		# if we're dealing with a peer node then we determine sameness
		# by the two nodes having the same server
		if self.type == NodeType.Peer:
			return self.ip == other.ip and self.port == other.port

		# if we're dealing with a resource node, we say two nodes are the same
		# if they have the same value
		return self.value == other.value

	def __iter__(self):
		return iter([self.digest_id, self.ip, self.port])

	def __hash__(self):
		return self.long_id

	def __repr__(self):
		return repr([self.long_id, self.ip, self.port])

	def __str__(self):
		# return f"<{self.type}, {self.ip}, {self.port}, {self.long_id}>"
		if self.type == NodeType.Peer:
			return f"{self.type}@{self.ip}:{self.port}"
		return f"{self.type}@{self.long_id}"


class NodeHeap:
	def __init__(self, node, maxsize):
		"""
		NodeHead

		A heaped binary tree featuring a set of neighbors ordered by distance
		via `node.distance_to()`. The heap can contain up maxsize nodes, and
		will return min(len(NodeHeap), maxsize) nodes from __iter__

		Parameters
		----------
			node: Node
				The node to measure all distnaces from.
			maxsize: int
				The maximum size that this heap can grow to.
		"""
		self.node = node
		self.heap = []
		self.contacted = set()
		self.maxsize = maxsize

	def remove(self, peers: List["Node"]) -> None:
		"""
		Remove a list of peer ids from this heap. Note that while this
		heap retains a constant visible size (based on the iterator), it's
		actual size may be quite a bit larger than what's exposed.  Therefore,
		removal of nodes may not change the visible size as previously added
		nodes suddenly become visible.

		Parameters
		----------
			peers: List[Node]
				List of peers which to prune
		"""
		peers = set(peers)
		if not peers:
			return
		nheap = []
		for distance, node in self.heap:
			if node.digest_id not in peers:
				heapq.heappush(nheap, (distance, node))
				continue
			log.debug("removing peer %s from node %s", str(node), str(node))
		self.heap = nheap

	def get_node(self, node_id):
		for _, node in self.heap:
			if node.digest_id == node_id:
				return node
		return None

	def have_contacted_all(self):
		return len(self.get_uncontacted()) == 0

	def get_ids(self):
		return [n.digest_id for n in self]

	def mark_contacted(self, node):
		self.contacted.add(node.digest_id)

	def popleft(self):
		return heapq.heappop(self.heap)[1] if self else None

	def push(self, nodes):
		"""
		Push nodes onto heap.

		@param nodes: This can be a single item or a C{list}.
		"""
		if not isinstance(nodes, list):
			nodes = [nodes]

		for node in nodes:
			if node not in self:
				distance = self.node.distance_to(node)
				heapq.heappush(self.heap, (distance, node))

	def __len__(self):
		return min(len(self.heap), self.maxsize)

	def __iter__(self):
		nodes = heapq.nsmallest(self.maxsize, self.heap)
		return iter(map(itemgetter(1), nodes))

	def __contains__(self, node):
		for _, other in self.heap:
			if node.digest_id == other.digest_id:
				return True
		return False

	def get_uncontacted(self):
		return [n for n in self if n.digest_id not in self.contacted]



import os
import time

from liaa.storage import EphemeralStorage, DiskStorage


class TestEphemeralStorage:
	# pylint: disable=no-self-use
	def test_can_instantiate_storage(self, mknode):
		storage = EphemeralStorage(mknode(), 10)
		assert isinstance(storage, EphemeralStorage)

	def test_can_set_and_get(self, mknode, mkrsrc):
		storage = EphemeralStorage(mknode, 10)
		resource = mkrsrc(key=b"one", value=b"two")
		storage.set(resource)

		node = storage.get(resource.hex)
		assert  node.value == b"two"

	def test_resource_expires_per_expiry(self, mknode, mkrsrc):
		# Expiry time of 0 should force the prune to make all items stale immediately
		storage = EphemeralStorage(mknode(), 0)
		resource = mkrsrc(key=b"one", value=b"two")
		storage.set(resource)
		assert not storage.get(resource.hex)

	def test_iter(self, mknode, mkrsrc):
		storage = EphemeralStorage(mknode(), 0)
		resource = mkrsrc(key=b"one", value=b"two")
		storage.set(resource)
		for node in storage:
			assert node.hex == resource.hex
			assert node.value == resource.value

	def test_iter_older_than_returns_proper_keys_for_republishing(self, mknode, mkrsrc):
		storage = EphemeralStorage(mknode(), 0)
		resource = mkrsrc(key=b"one", value=b"two")
		storage.set(resource)
		for key, value in storage.iter_older_than(0):
			assert key == resource.hex
			assert value == resource.value


class TestDiskStorage:

	# pylint: disable=no-self-use
	def test_instantiation(self, mknode):
		node = mknode()
		storage = DiskStorage(node)

		assert isinstance(storage, DiskStorage)
		assert len(storage) == 0

	def test_store_contents(self, mknode, mkrsrc):
		node = mknode()
		storage = DiskStorage(node)

		resource = mkrsrc()
		storage.set(resource)

		assert len(storage) == len(storage.contents()) == 1
		assert storage.contents()[0] == str(resource.hex)

	def test_remove_works_ok(self, mknode, mkrsrc):
		node = mknode()
		storage = DiskStorage(node)

		resource = mkrsrc()
		storage.set(resource)

		assert len(storage) == 1

		storage.remove(resource.hex)
		assert len(storage) == 0

	def test_load_data_returns_proper_payload(self, mknode, mkrsrc):
		node = mknode()
		storage = DiskStorage(node)

		resource = mkrsrc()
		storage.set(resource)

		# pylint: disable=protected-access
		data = storage._load_data(resource.hex)
		assert resource.value == data

	def test_persist_dir_exists(self, mknode):
		node = mknode()
		storage = DiskStorage(node)

		assert os.path.exists(storage.dir)
		assert os.path.exists(storage.content_dir)

	def test_can_set_and_retrieve_basic_resource(self, mknode, mkrsrc):
		node = mknode()
		storage = DiskStorage(node)
		resource = mkrsrc()

		storage.set(resource)

		result = storage.get(resource.hex)
		assert result == resource


	def test_prune_removes_old_data(self, mknode, mkrsrc):
		node = mknode()

		# instantiate disk storage with 5 seconds ttl
		storage = DiskStorage(node, ttl=3)

		# create some resources
		# pylint: disable=invalid-name
		r1 = mkrsrc()
		r2 = mkrsrc()

		# set our first node
		storage.set(r1)

		# sleep over 3 seconds
		time.sleep(4)

		# our next call to set will call storage.prune thus removing r1
		storage.set(r2)

		# we should only have r2
		assert len(storage) == 1

		for node in storage:
			assert node.hex == r2.hex
			assert node.value == r2.value



import os
import random
import hashlib
import struct
import umsgpack

import pytest

# pylint: disable=bad-continuation
from liaa.protocol import Header
from liaa.network import Server
from liaa.node import Node, NodeType
from liaa.protocol import KademliaProtocol
from liaa.routing import RoutingTable, KBucket
from liaa.storage import StorageIface
from liaa.utils import rand_digest_id, rand_str


@pytest.yield_fixture
def bootstrap_node(event_loop):
	server = Server()
	event_loop.run_until_complete(server.listen(8468))

	try:
		yield ('127.0.0.1', 8468)
	finally:
		server.stop()

# pylint: disable=redefined-outer-name
@pytest.fixture()
def mknode():
	# pylint: disable=invalid-name
	def _mknode(digest_id=None, ip=None, port=None, intid=None):
		"""
		Make a node.  Created a random id if not specified.
		"""
		if intid is not None:
			digest_id = struct.pack('>l', intid)
		if not digest_id:
			randbits = str(random.getrandbits(255))
			digest_id = hashlib.sha1(randbits.encode()).digest()
		return Node(digest_id, ip, port)
	return _mknode


@pytest.fixture()
def mkdgram():
	def _mkdgram(header=Header.Request, msg_id=os.urandom(32), data=('funcname', 123)):
		"""
		Create a datagram
		"""
		return header + hashlib.sha1(msg_id).digest() + umsgpack.packb(data)
	return _mkdgram


@pytest.fixture()
def mkrsrc():
	def _mkrsrc(key=None, value=None):
		"""
		Create a fake resource
		"""
		key = key or rand_digest_id()
		value = value or rand_str()
		# pylint: disable=bad-continuation
		return Node(digest_id=key,
						type=NodeType.Resource,
						value=value)
	return _mkrsrc


# pylint: disable=too-few-public-methods
@pytest.fixture()
def mkbucket():
	def _mkbucket(ksize, low=0, high=2**160):
		return KBucket(low, high, ksize)
	return _mkbucket


# pylint: disable=too-few-public-methods
class FakeProtocol(KademliaProtocol):  # pylint: disable=too-few-public-methods
	def __init__(self, source_id, storage, ksize=20):
		super(FakeProtocol, self).__init__(source_id, storage=storage, ksize=ksize)
		self.router = RoutingTable(self, ksize, Node(source_id))
		self.source_id = source_id


@pytest.fixture()
def fake_proto(mknode):
	def _fake_proto(node=None):
		node = node or mknode()
		return FakeProtocol(node.digest_id, StorageIface(node), ksize=20)
	return _fake_proto


# pylint: disable=too-few-public-methods
class FakeServer:
	def __init__(self, node):
		self.node_id = node.digest_id
		self.storage = StorageIface(node)
		self.ksize = 20
		self.alpha = 3
		self.protocol = FakeProtocol(self.node_id, self.storage, self.ksize)
		self.router = self.protocol.router


@pytest.fixture
def fake_server(mknode):
	return FakeServer(mknode())


class Sandbox:
	def __init__(self, obj):
		self.obj = obj
		self.mem = {}

	def stub(self, funcname, func):
		self.mem[funcname] = getattr(self.obj, funcname)
		setattr(self.obj, funcname, func)

	def restore(self):
		for funcname, func in self.mem.items():
			setattr(self.obj, funcname, func)


@pytest.fixture
def sandbox():
	def _sandbox(obj=None):
		if not obj:
			raise RuntimeError("sandbox object cannot be None")
		return Sandbox(obj)
	return _sandbox



import pytest

from liaa.crawling import SpiderCrawl
from liaa.network import KademliaProtocol
from liaa.node import NodeHeap


# pylint: disable=too-few-public-methods
class FakeSpiderCrawler:
	def __init__(self, protocol, node, peers, ksize, alpha):
		self.protocol = protocol
		self.node = node
		self.peers = peers
		self.ksize = ksize
		self.alpha = alpha

def fake_spider_crawler(node):
	return SpiderCrawl(KademliaProtocol, node=node, peers=[], ksize=3, alpha=3)

class TestSpiderCrawl:
	# pylint: disable=no-self-use
	def test_can_init_crawler(self, mknode):
		crawler = fake_spider_crawler(node=mknode())
		assert isinstance(crawler, SpiderCrawl)
		assert isinstance(crawler.nearest, NodeHeap)
		assert len(crawler.nearest) == 0

	@pytest.mark.skip(reason="not implemented")
	def test_find_returns_expected_result(self, mknode):
		crawler = fake_spider_crawler(node=mknode())
		assert len(crawler.last_ids_crawled) == 0
		crawler.nearest.push([mknode(intid=i) for i in range(5)])



import asyncio

# pylint: disable=bad-continuation
from liaa.utils import (
	shared_prefix,
	bytes_to_bit_string,
	hex_to_int,
	ArgsParser,
	check_dht_value_type,
	gather_dict,
	join_addr,
	hex_to_int_digest,
	split_addr,
	digest_to_int,
	rand_str,
	rand_int_id,
	int_to_digest
)


class TestUtils:

	# pylint: disable=no-self-use
	def test_shared_prefix(self):
		args = ['prefix', 'prefixasdf', 'prefix', 'prefixxxx']
		assert shared_prefix(args) == 'prefix'

		args = ['p', 'prefixasdf', 'prefix', 'prefixxxx']
		assert shared_prefix(args) == 'p'

		args = ['one', 'two']
		assert shared_prefix(args) == ''

		args = ['hi']
		assert shared_prefix(args) == 'hi'

	def test_bytes_to_bit_string(self):
		arr = b"hello world"
		bstr = [bin(bite)[2:].rjust(8, '0') for bite in arr]
		assert "".join(bstr) == bytes_to_bit_string(arr)

	def test_to_base16_int(self):
		num = 5
		num_as_bytes = bytes([num])
		assert hex_to_int(num_as_bytes.hex()) == num

	def test_check_dht_value_type_returns_true_when_arg_is_valid(self):
		# pylint: disable=invalid-name
		a = check_dht_value_type("foo")
		assert a

		b = check_dht_value_type(8)
		assert b

		c = check_dht_value_type(b'123')
		assert c

		d = check_dht_value_type(True)
		assert d

		e = check_dht_value_type(3.14)
		assert e

		f = check_dht_value_type({})
		assert not f

		g = check_dht_value_type([])
		assert not g

		# pylint: disable=too-few-public-methods
		class Foo:
			pass

		h = check_dht_value_type(Foo())
		assert not h

	def test_gather_dict(self):
		coros = [asyncio.sleep(1) for _ in range(3)]
		keys = [str(i) for i in range(3)]
		coros = dict(zip(keys, coros))

		loop = asyncio.new_event_loop()
		results = loop.run_until_complete(gather_dict(coros))
		assert isinstance(results, dict)
		assert len(results) == 3

		assert all(map(lambda p: not p[1], results.items()))

	def test_join_addr(self):
		addr = ("0.0.0.0", 8000)
		assert join_addr(addr) == "0.0.0.0:8000"

	def test_split_addr(self):
		addr = "0.0.0.0:8000"
		assert split_addr(addr) == ("0.0.0.0", 8000)

	def test_rand_str(self):
		result = rand_str()
		assert isinstance(result, str)
		assert len(result) == 20

	def test_rand_int_id(self):
		result = rand_int_id()
		assert isinstance(result, int)
		assert 0 < result < 2**160

	def test_digest_to_int(self):
		num = 10
		byte_arr = num.to_bytes(16, byteorder='big')
		assert digest_to_int(byte_arr) == num

	def test_int_to_digest(self):
		num = 10
		byte_arr = num.to_bytes((num.bit_length() // 8) + 1, byteorder='big')
		assert byte_arr == int_to_digest(num)

	def digest_to_int(self):
		dgst = b'\x00\x00\x00\n'
		assert dgst == digest_to_int(10)

	def test_hex_to_int_digest(self):
		num = 10
		hexval = int_to_digest(num).hex()
		assert hex_to_int_digest(hexval) == int_to_digest(num)

class TestArgsParser:

	# pylint: disable=no-self-use
	def test_can_set_and_get_via_primary_flag(self):
		parser = ArgsParser()
		opts = [("-f", "foo"), ("--bar", "bar"), ("-z", "1")]
		parser.add_many(opts)

		assert parser.get("-f", "--foo") == "foo"

	def test_can_set_and_get_via_secondary_flag(self):
		parser = ArgsParser()
		opts = [("--foo", "foo"), ("--bar", "bar"), ("-z", "1")]
		parser.add_many(opts)

		assert parser.get("-f", "--foo") == "foo"

	def test_has_help_opt_is_true(self):
		parser = ArgsParser()
		opts = [("-h", "")]
		parser.add_many(opts)
		assert parser.has_help_opt()



import random
import hashlib
from collections.abc import Iterable

from liaa.node import Node, NodeHeap
from liaa.utils import hex_to_int


class TestNode:
	# pylint: disable=no-self-use
	def test_node_long_id_derivation_is_ok(self):
		rid = hashlib.sha1(str(random.getrandbits(255)).encode()).digest()
		node = Node(rid)
		assert node.long_id == hex_to_int(rid.hex())

	def test_valid_diff_calculation_is_ok(self):
		ridone = hashlib.sha1(str(random.getrandbits(255)).encode())
		ridtwo = hashlib.sha1(str(random.getrandbits(255)).encode())

		shouldbe = hex_to_int(ridone.digest().hex()) ^ hex_to_int(ridtwo.digest().hex())
		none = Node(ridone.digest())
		ntwo = Node(ridtwo.digest())
		assert none.distance_to(ntwo) == shouldbe

	def test_distance_diff_of_same_node_is_0(self, mknode):
		node_one = mknode(intid=1)
		node_two = mknode(intid=1)
		assert node_one.distance_to(node_two) == 0

	def test_is_same_node(self, mknode):
		node_one = mknode(ip="0.0.0.0", port=0000)
		node_two = mknode(ip="0.0.0.0", port=0000)
		assert node_one.is_same_node(node_two)

	def test_node_iter(self, mknode):
		node = mknode()
		assert tuple(node) == (node.digest_id, node.ip, node.port)


class TestNodeHeap:
	# pylint: disable=no-self-use
	def test_can_create_nodeheap(self, mknode):
		heap = NodeHeap(mknode(), 2)
		assert isinstance(heap, NodeHeap)

	def test_get_node_return_node_when_node_present(self, mknode):
		heap = NodeHeap(mknode(), 3)
		nodes = [mknode(intid=i) for i in range(3)]
		# pylint: disable=invalid-name
		for n in nodes:
			heap.push(n)
		node = heap.get_node(nodes[0].digest_id)
		assert isinstance(node, Node)

	def test_get_node_returns_none_when_node_not_exists(self, mknode):
		heap = NodeHeap(mknode(), 1)
		empty = heap.get_node(123)
		assert not empty

	def test_mark_contacted_works_ok(self, mknode):
		maxsize = 10
		heap = NodeHeap(mknode(), maxsize)
		nodes = [mknode(intid=i) for i in range(maxsize)]
		for node in nodes:
			heap.push(node)
		contacted = nodes[:5]
		for node in contacted:
			heap.mark_contacted(node)

		assert len(heap.contacted) == 5
		assert not heap.have_contacted_all()
		assert len(heap.get_uncontacted()) == 5

	def test_popleft_returns_left_if_heap_not_empty(self, mknode):
		maxsize = 5
		heap = NodeHeap(mknode(), maxsize)
		nodes = [mknode(intid=i) for i in range(maxsize)]
		for node in nodes:
			heap.push(node)

		popped = heap.popleft()
		assert isinstance(popped, Node)

	def test_popleft_returns_none_when_heap_empty(self, mknode):
		maxsize = 1
		heap = NodeHeap(mknode(), maxsize)
		nodes = [mknode(intid=i) for i in range(maxsize)]
		for node in nodes:
			heap.push(node)

		heap.remove(nodes)
		popped = heap.popleft()
		assert isinstance(popped, Iterable)
		# pylint: disable=invalid-name
		_, ip, port = tuple(popped)
		assert (not ip) and (not port)


	def test_heap_overload_doesnt_exceed_maxsize(self, mknode):
		maxsize = 3
		node = NodeHeap(mknode(intid=0), maxsize)
		assert not node

		for digit in range(10):
			node.push(mknode(intid=digit))

		assert len(node) == maxsize
		assert len(list(node)) == maxsize

	def test_heap_iters_over_nsmallest_via_distance(self, mknode):
		heap = NodeHeap(mknode(intid=0), 5)
		nodes = [mknode(intid=x) for x in range(10)]
		for index, node in enumerate(nodes):
			heap.push(node)

		for index, node in enumerate(heap):
			assert index == node.long_id
			assert index < 5

	def test_remove(self, mknode):
		maxsize = 5
		heap = NodeHeap(mknode(intid=0), maxsize)
		nodes = [mknode(intid=x) for x in range(10)]
		for node in nodes:
			heap.push(node)

		heap.remove([nodes[0].digest_id, nodes[1].digest_id])
		# len(heap) == min(10-2, maxsize)
		assert len(list(heap)) == maxsize

		for index, node in enumerate(heap):
			# we removed to elements so offset index to account for it
			assert index + 2 == node.long_id
			assert index < maxsize



import asyncio
import logging
import os
import pickle
from typing import List, Optional, Tuple

from liaa.crawling import NodeSpiderCrawl, ValueSpiderCrawl
from liaa.node import Node, NodeType
from liaa.protocol import KademliaProtocol, HttpInterface
from liaa.storage import StorageIface
from liaa.utils import int_to_digest, rand_digest_id

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


# pylint: disable=too-many-instance-attributes
class Server:

	protocol_class = KademliaProtocol

	# pylint: disable=bad-continuation
	def __init__(self,
		node_id: Optional[bytes] = None, ksize: int = 20, alpha: int = 3, **kwargs):
		"""
		High level view of a node instance.  This is the object that should be
		created to start listening as an active node on the network.
		Create a server instance.  This will start listening on the given port.

		Parameters
		----------
			node_id: Optional[bytes]
				The id for this node on the network.
			ksize: int
				The k parameter from the paper (default = 20)
			alpha: int
				The alpha parameter from the paper (default = 3)
		"""
		self.node = Node(node_id if node_id else rand_digest_id())
		log.info("Using storage interface: %s", StorageIface.__name__)
		self.storage = StorageIface(self.node)
		self.ksize = ksize
		self.alpha = alpha
		self.refresh_interval = kwargs.get("refresh_interval")
		self.statefile = os.path.join(self.storage.dir, "node.state")

		self.udp_transport = None
		self.protocol = None
		self.refresh_loop = None
		self.save_state_loop = None
		self.listener = None

	def stop(self) -> None:
		"""
		Stop a currently running server - all event loops, and close
		all open network connections
		"""
		if self.udp_transport is not None:
			log.info("Closing %s udp transport...", self.node)
			self.udp_transport.close()

		if self.refresh_loop:
			log.info("Cancelling %s refresh loop...", self.node)
			self.refresh_loop.cancel()

		if self.save_state_loop:
			log.info("Cancelling %s state loop...", self.node)
			self.save_state_loop.cancel()

		if self.listener:
			log.info("Closing %s server...", self.node)
			asyncio.ensure_future(self.listener.wait_closed())

	def _create_protocol(self) -> "KademliaProtocol":
		"""
		Create an instance of the Kademlia protocol

		Returns
		-------
			KademliaProtocol:
				Instance of the kademlia protocol
		"""
		return self.protocol_class(self.node, self.storage, self.ksize)

	def _create_http_iface(self) -> "HttpInterface":
		"""
		Create an interface to accept incoming http messages

		Returns
		-------
			HttpInterface:
				Bootstrapped instance of an HttpInterface
		"""
		return HttpInterface(self.node, self.storage)

	async def listen(self, port: int, interface: str = "0.0.0.0") -> None:
		"""
		Create UDP and HTTP listeners on an interface at a given port

		Parameters
		----------
			port: int
				Port on which to bind inteface
			interface: str
				Interface on which to listen (default = 0.0.0.0)
		"""
		self.node.ip = interface
		self.node.port = port

		loop = asyncio.get_event_loop()
		# pylint: disable=bad-continuation
		listen = loop.create_datagram_endpoint(self._create_protocol,
												local_addr=(interface, port))
		log.info("%s UDP listening at %s:%i", self.node, interface, port)

		self.udp_transport, self.protocol = await listen

		# schedule refreshing table
		self.refresh_table()

		# pylint: disable=bad-continuation
		self.listener = await loop.create_server(self._create_http_iface,
												host=interface, port=port)
		log.info("%s HTTP listening at %s:%i", self.node, interface, port)

		asyncio.ensure_future(self.listener.serve_forever())

	def refresh_table(self) -> None:
		"""
		Refresh our routing table and save our server's state
		"""
		interval = self.refresh_interval or 10
		log.debug("Refreshing routing table for %s", self.node)
		asyncio.ensure_future(self._refresh_table())
		loop = asyncio.get_event_loop()
		self.refresh_loop = loop.call_later(interval, self.refresh_table)
		self.save_state_loop = loop.call_later(interval, self.save_state_regularly)

	async def _refresh_table(self) -> None:
		"""
		Refresh buckets that haven't had any lookups in the last hour
		(per section 2.3 of the paper).
		"""
		results: List[asyncio.Future] = []
		for digest_id in self.protocol.get_refresh_ids():
			node = Node(digest_id)
			nearest = self.protocol.router.find_neighbors(node, self.alpha)
			log.debug("%s refreshing routing table on %i nearest", self.node, len(nearest))
			spider = NodeSpiderCrawl(self.protocol, node, nearest, self.ksize, self.alpha)
			results.append(spider.find())

		# do our crawling
		await asyncio.gather(*results)

		# now republish keys older than one hour
		for node in self.storage.iter_older_than(3600):
			# node = Node(hex_to_int_digest(hexkey), type=NodeType.Resource, value=value)
			log.debug("%s republishing node %s from store", self.node, node)
			await self.set_digest(node)

	def bootstrappable_neighbors(self) -> List["Node"]:
		"""
		Get a list of (ip, port) tuple pairs suitable for use as an argument to
		the bootstrap method.

		The server should have been bootstrapped
		already - this is just a utility for getting some neighbors and then
		storing them if this server is going down for a while.  When it comes
		back up, the list of nodes can be used to bootstrap.

		Returns
		-------
			List[Node]:
				List of peers suitable for bootstrap use
		"""
		neighbors: List["Node"] = self.protocol.router.find_neighbors(self.node)
		return [tuple(n)[-2:] for n in neighbors]

	async def bootstrap(self, addrs: List[Tuple[str, int]]) -> asyncio.Future:
		"""
		Bootstrap the server by connecting to other known nodes in the network.

		Parameters
		----------
			addrs: List[Tuple[str, int]]
				Note that only IP addresses are acceptable - hostnames will
				cause an error.

		Returns
		-------
			asyncio.Future:
				scheduled callback for a NodeSpiderCrawl to continue crawling
				network in order to find peers for self.node
		"""
		log.debug("%s attempting to bootstrap with contacts: %s", self.node, addrs)
		cos = list(map(self.bootstrap_node, addrs))
		gathered = await asyncio.gather(*cos)
		nodes = [node for node in gathered if node is not None]
		spider = NodeSpiderCrawl(self.protocol, self.node, nodes, self.ksize, self.alpha)
		return await spider.find()

	async def bootstrap_node(self, addr: Tuple[str, int]) -> Optional["Node"]:
		"""
		Ping a given address so that both `addr` and `self.node` can know
		about one another

		Parameters
		----------
			addr: Tuple[str, int]
				Address of peer to ping

		Returns
		-------
			Optiona[Node]:
				None if ping was unsuccessful, or peer as Node if ping
				was successful
		"""
		result = await self.protocol.ping(addr, self.node.digest_id)
		return Node(result[1], addr[0], addr[1]) if result[0] else None

	async def get(self, key: int) -> asyncio.Future:
		"""
		Crawl the current node's known network in order to find a given key. This
		is the interface for grabbing a key from the network

		Parameters
		----------
			key: int
				Key to find in network

		Returns
		-------
			asyncio.Future:
				A recursive call to ValueSpiderCrawl.find which will terminate
				either when the value is find or the search is exhausted
		"""
		log.info("%s looking up key %s", self.node, key)
		dkey = int_to_digest(key)

		# if this node has it, return it
		result = self.storage.get(dkey.hex())
		if result is not None:
			return result

		node = Node(dkey, type=NodeType.Resource)
		nearest = self.protocol.router.find_neighbors(node)

		if not nearest:
			log.warning("There are no known neighbors to get key %s", str(node))
			return None

		spider = ValueSpiderCrawl(self.protocol, node, nearest, self.ksize, self.alpha)
		return await spider.find()

	async def set(self, node: "Node") -> asyncio.Future:
		"""
		Set the given string key to the given value in the network. This is the
		interface for setting a key throughout the network

		Parameters
		----------
			node: Node
				Node to be stored

		Returns
		-------
			asyncio.Future:
				Callback to set_digest which finds nodes on which to store key
		"""
		if not node.has_valid_value():
			raise TypeError("Value must be of type int, float, bool, str, or bytes")
		log.info("setting '%s' = '%s' on network", str(node), node.value)
		return await self.set_digest(node)

	async def set_digest(self, node: "Node") -> bool:
		"""
		Set the given SHA1 digest key (bytes) to the given value in the
		network.

		Parameters
		----------
			node: Node
				Node to be set (where node.type == NodeType.Resource)

		Returns
		-------
			bool:
				Indicator of whether or not the key/value pair was stored
				on any of the nearest nodes found by the SpiderCrawler
		"""
		nearest = self.protocol.router.find_neighbors(node)
		if not nearest:
			log.warning("There are no known neighbors to set key %s", node.hex)
			return False

		spider = NodeSpiderCrawl(self.protocol, node, nearest, self.ksize, self.alpha)
		nodes = await spider.find()
		log.info("setting '%s' on %s", str(node), ",".join(list(map(str, nodes))))

		# if this node is close too, then store here as well
		biggest = max([n.distance_to(node) for n in nodes])
		if self.node.distance_to(node) < biggest:
			self.storage.set(node)
		results = [self.protocol.call_store(n, node.digest_id, node.value) for n in nodes]

		# return true only if at least one store call succeeded
		return any(await asyncio.gather(*results))

	def save_state(self, fname: Optional[str] = None) -> None:
		"""
		Save the state of this node (the alpha/ksize/id/immediate neighbors)
		to a cache file with the given fname.

		Parameters
		----------
			fname: strtest_can_set_and_get
				File location where in which to write state
		"""
		fname = fname or self.statefile
		log.info("%s saving state to %s", self.node, fname)

		# pylint: disable=bad-continuation
		data = {
			'ksize': self.ksize,
			'alpha': self.alpha,
			'id': self.node.digest_id,
			'neighbors': self.bootstrappable_neighbors()
		}

		if not data['neighbors']:
			log.warning("%s has no known neighbors, so not writing to cache.", self.node)
			return

		with open(fname, 'wb') as file:
			pickle.dump(data, file)

	def load_state(self, fname: Optional[str] = None) -> "Server":
		"""
		Load the state of this node (the alpha/ksize/id/immediate neighbors)
		from a cache file with the given fname.

		Parameters
		----------
			fname: Optional[str]
				File location where in which to write state
					(default is node.state file in storage directory)

		Returns
		-------
			Server:
				Parameterized instance of a Server
		"""
		fname = fname or self.statefile
		log.info("%i loading state from %s", self.node, fname)

		with open(fname, 'rb') as file:
			data = pickle.load(file)

		svr = Server(data['id'], ksize=data['ksize'], alpha=data['alpha'])

		if data['neighbors']:
			asyncio.ensure_future(svr.bootstrap(data['neighbors']))

		return svr

	# pylint: disable=bad-continuation
	def save_state_regularly(self, fname: Optional[str] = None, frequency: int = 600) -> None:
		"""
		Save the state of node with a given regularity to the given
		filename.

		Parameters
		----------
			fname: Optional[str]
				Location at which to save state regularly
					(default is node.state file in storage directory)
			frequency: int
				Frequency in seconds that the state should be saved. (default=10 mins)
		"""
		fname = fname or self.statefile
		self.save_state(fname)
		loop = asyncio.get_event_loop()

		# pylint: disable=bad-continuation
		self.save_state_loop = loop.call_later(frequency,
												self.save_state_regularly,
												fname,
												frequency)



import asyncio
import base64
import hashlib
import logging
import sys
import os
import json
import random
from typing import Any, Dict, List, Tuple, Union, Optional


import umsgpack


from liaa.node import Node, NodeType
from liaa.routing import RoutingTable
from liaa import __version__
from liaa.utils import join_addr, rand_digest_id

log = logging.getLogger(__name__)  # pylint: disable=invalid-name

# pylint: disable=too-few-public-methods
class Header:
	Request = b'\x00'
	Response = b'\x01'


class MalformedMessage(Exception):
	pass


class RPCFindResponse:
	def __init__(self, response: List[Dict[str, Any]]):
		"""
		RPCFindResponse

		A wrapper for the result of a RPC find.

		Parameters
		----------
			response: Tuple[bool, Union[List[Tuple[int, str, int]], Dict[str, Any]]]
				This will be a tuple of (<response received>, <value>)
				where <value> will be a list of tuples if not found or
				a dictionary of {'value': v} where v is the value desired
		"""
		self.response = response

	def did_happen(self) -> bool:
		"""
		Did the other host actually respond?

		Returns
		-------
			bool:
				Indicator of host response
		"""
		return self.response[0]

	def has_value(self) -> bool:
		"""
		Return whether or not the response has a value

		Returns
		-------
			bool:
				Whether or not the value of the response contains a found value
		"""
		return isinstance(self.response[1], dict)

	def get_value(self) -> Any:
		"""
		Get the value/payload from a response that contains a value

		Returns
		-------
			Any:
				Value of the response payload
		"""
		return self.response[1]['value']

	def get_node_list(self) -> List["Node"]:
		"""
		Get the node list in the response.  If there's no value, this should
		be set.

		Returns
		-------
			List["Node"]:
				List of nodes returned from find response
		"""
		nodelist = self.response[1] or []
		return [Node(*nodeple) for nodeple in nodelist]


class RPCDatagramProtocol(asyncio.DatagramProtocol):
	def __init__(self, wait: int = 5):
		"""
		RPCDatagramProtocol

		Protocol implementation using msgpack to encode messages and asyncio
		to handle async sending / recieving.

		Parameters
		----------
			wait: int
				Network connection timeout (default=5)
		"""
		self._wait = wait
		self._outstanding_msgs: Dict[int, Tuple[asyncio.Future, asyncio.Handle]] = {}
		self._new_queue = {}
		self.transport = None

	def connection_made(self, transport: asyncio.Handle) -> None:
		"""
		Called when a connection is made. (overload from BaseProtocol)

		Parameters
		----------
			transport: asyncio.Handle
				The transport representing the connection. The protocol is
				responsible for storing the reference to its transport
		"""
		self.transport = transport

	def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
		"""
		Called when a datagram is received.

		Parameters
		----------
			data: bytes
				object containing the incoming data.
			addr: Tuple
				address of the peer sending the data; the exact format depends on the transport.
		"""
		log.debug("incoming dgram from peer at %s", join_addr(addr))
		asyncio.ensure_future(self._solve_dgram(data, addr))

	async def _solve_dgram(self, buff: bytes, address: Tuple[str, int]) -> None:
		"""
		Responsible for processing an incoming datagram

		Parameters
		----------
			buff: bytes
				Data to be processed
			address: Tuple
				Address of sending peer
		"""
		if len(buff) < 22:
			log.warning("received invalid dgram from %s, ignoring", address)
			return

		if buff[:1] == Header.Request:
			asyncio.ensure_future(self._accept_request(buff, address))
		elif buff[:1] == Header.Response:
			self._accept_response(buff, address)
		else:
			log.debug("Received unknown message from %s, ignoring", address)

	def _accept_response(self, buff: bytes, address: Tuple[str, int]) -> None:
		"""
		Processor for incoming responses

		Parameters
		----------
			buff: bytes
				Datagram representing incoming message from peer
			address: Tuple
				Address of peer receiving response
		"""
		idf, data = buff[1:21], umsgpack.unpackb(buff[21:])
		msgargs = (base64.b64encode(idf), address)
		if idf not in self._outstanding_msgs:
			log.warning("received unknown message %s from %s; ignoring", *msgargs)
			return

		# pylint: disable=bad-continuation
		log.debug("received %iB response for message id %s from %s", sys.getsizeof(data),
					idf, join_addr(msgargs[1]))

		if not idf in self._new_queue:
			log.warning("could not mark datagram %s as received", idf)
			return

		fut, timeout = self._new_queue[idf]
		fut.set_result((True, data))
		timeout.cancel()
		del self._new_queue[idf]

	async def _accept_request(self, buff: bytes, address: Tuple[str, int]) -> None:
		"""
		Process an incoming request datagram as well as its RPC response

		Parameters
		----------
			buff: bytes
				Datagram representing incoming message from peer
			address: Tuple
				Address of sender
		"""
		idf, data = buff[1:21], umsgpack.unpackb(buff[21:])
		funcname, args = data

		# basically, when we execute an operation such as protocol.ping(), we will
		# send 'ping' as the function name, at which point, we concat 'rpc_' onto
		# the function name so that when we call getattr(self, funcname) we will
		# get the rpc version of the fucname

		func = getattr(self, "rpc_%s" % funcname, None)
		if func is None or not callable(func):
			msgargs = (self.__class__.__name__, funcname)
			log.warning("%s has no callable method rpc_%s; ignoring request", *msgargs)
			return

		if not asyncio.iscoroutinefunction(func):
			func = asyncio.coroutine(func)

		response = await func(address, *args)
		txdata = Header.Response + idf + umsgpack.packb(response)
		# pylint: disable=bad-continuation
		log.debug("sending %iB response for msg id %s to %s", len(txdata),
			base64.b64encode(idf), join_addr(address))
		self.transport.sendto(txdata, address)

	def _timeout(self, msg_id: int) -> None:
		"""
		Make a given datagram timeout

		Parameters
		----------
			msg_id: int
				ID of datagram future to cancel
		"""
		args = (base64.b64encode(msg_id), self._wait)
		log.error("Did not received reply for msg id %s within %i seconds", *args)
		self._outstanding_msgs[msg_id][0].set_result((False, None))
		del self._outstanding_msgs[msg_id]

	def __getattr__(self, name: str) -> Union[asyncio.Future, asyncio.Future, None]:

		if name.startswith("_") or name.startswith("rpc_"):
			return getattr(super(), name)

		try:
			# else we follow normal getattr behavior (same as above)
			return getattr(super(), name)
		except AttributeError:
			pass

		# here we define a catchall function that creates a request using a given
		# function name and *args. these *args are sent and pushed to the msg queue
		# as futures. this closure being called means that we are trying to execute
		# a function name that is not part of the base Kademlia rpc_* protocol
		def func(address, *args):
			msg_id = hashlib.sha1(os.urandom(32)).digest()
			data = umsgpack.packb([name, args])
			if len(data) > 8192:
				raise MalformedMessage("Total length of function name and arguments cannot exceed 8K")
			txdata = Header.Request + msg_id + data

			# pylint: disable=bad-continuation
			log.debug("Attempting to execute rpc %s on peer at %s - msgid %s",
						name, join_addr(address), base64.b64encode(msg_id))
			self.transport.sendto(txdata, address)

			# we assume python version >= 3.7
			loop = asyncio.get_event_loop()
			future = loop.create_future()
			timeout = loop.call_later(self._wait, self._timeout, msg_id)
			# self._queue.enqueue_fut(msg_id, future, timeout)
			self._new_queue[msg_id] = (future, timeout)
			self._outstanding_msgs[msg_id] = (future, timeout)
			return future

		return func


class HttpInterface(asyncio.Protocol):
	def __init__(self, source_node: "Node", storage: "IStorage", wait: int = 5):
		"""
		HttpInterface

		Public interface for storing data and fetching data on this server

		Parameters
		----------
			source_node: Node
				Our node (representing the current machine)
			storage: IStorage
				Storage interface
			wait: int
				Timeout of requests
		"""
		self.source_node = source_node
		self.storage = storage
		self.wait = wait
		self.transport = None

	def connection_made(self, transport: asyncio.Handle) -> None:
		"""
		Called when a connection is made. (overload from BaseProtocol)

		Parameters
		----------
			transport: asyncio.Handle
				The transport representing the connection. The protocol is
				responsible for storing the reference to its transport
		"""
		self.transport = transport

	def data_received(self, data: bytes) -> None:
		"""
		Called when in an incoming stream is received.

		Parameters
		----------
			data: bytes
				object containing the incoming data.
		"""
		asyncio.ensure_future(self._handle_data(data))

	async def _handle_data(self, buff: bytes) -> None:
		"""
		Process the incoming stream

		Parameters
		----------
			buff: bytes
				The buffered stream to be process
		"""
		data = buff.decode()
		rawheaders, rawbody = data.split("\r\n\r\n")
		unmarshalled = json.loads(rawbody)

		log.debug("Received new http message %s", data)

		if rawheaders.startswith("GET"):
			response = self.fetch_data(unmarshalled.get("key"))
		elif rawheaders.startswith("PUT"):
			response = self.call_store(rawbody)
		else:
			body = json.dumps({"details": "Not implemented"})
			response = self.pack_response(501, "NOT IMPLEMENTED", body)

		self.transport.write(response.encode())
		self.transport.close()

	def call_store(self, payload: Optional[Any]) -> str:
		"""
		Given a payload, save it to storage

		Parameters
		----------
			payload: Optional[Any]
				Data to be saved

		Returns
		-------
			str:
				Response to write to client
		"""
		node = Node(rand_digest_id(), type=NodeType.Resource, value=payload)
		self.storage.set(node)
		return self.pack_response(200, "OK", json.dumps({"details": "ok"}))

	def fetch_data(self, key: Optional[str]) -> str:
		"""
		Given a key, return the its value, if we have the key in storage

		Parameters
		----------
			key: Optional[str]
				Hexkey of the resource to be found

		Returns
		-------
			str:
				Response to write to client
		"""
		node = self.storage.get(key)
		# pylint: disable=bad-continuation
		if node:
			return self.pack_response(200, "OK",
					json.dumps({"details": "found", "data": str(node)}))
		return self.pack_response(404, "NOT FOUND",
				json.dumps({"details": "Not found", "data": str(node)}))

	# pylint: disable=no-self-use
	def pack_response(self, code: int, msg: str, body: Dict[str, str]) -> str:
		"""
		Pack a response's parts into an HTTP message

		Parameters
		----------
			code: int
				Response status code
			msg: str
				Response status message
			body: Dict[str, str]
				Marshalled body payload

		Returns
		-------
			str:
				String-formatted response
		"""
		# pylint: disable=bad-continuation
		headers = [
			f"HTTP/1.1 {msg} {code}",
			f"User-Agent: Liaa.{__version__}",
			"Accept: */*",
			f"Content-Length: {len(body)}",
			f"Content-Type: application/x-www-form-urlencoded",
		]

		return "\r\n".join(headers) + "\r\n\r\n" + body

class KademliaProtocol(RPCDatagramProtocol):
	# pylint: disable=no-self-use,bad-continuation
	def __init__(self,
		source_node: "Node",
		storage: "EphemeralStorage",
		ksize: int
	):
		"""
		KadmeliaProtocol

		Abstraction used as a layer between our router and storage, and our
		public server. Protocol is responsible for executing various rpc's
		in order to update routing table, storage and keep network active

		Parameters
		----------
			source_node: Node
				Our node (representing the current machine)
			storage: IStorage
				Storage interface
			ksize: int
				Size of kbuckets
		"""
		RPCDatagramProtocol.__init__(self)
		self.router = RoutingTable(self, ksize, source_node)
		self.storage = storage
		self.source_node = source_node

	def get_refresh_ids(self):
		"""
		Get list of node ids with which to search, in order to keep old
		buckets up to date.

		Returns
		-------
			ids: List[int]
				ids of buckets that have not been updated since 3600
		"""
		ids = []
		for bucket in self.router.lonely_buckets():
			rid = random.randint(*bucket.range).to_bytes(20, byteorder='big')
			ids.append(rid)
		return ids

	def rpc_stun(self, sender: "Node") -> "Node":
		"""
		Execute a S.T.U.N procedure on a given sender

		Parameters
		----------
			sender: Node
				Requesting node

		Returns
		-------
			sender: Node
				Requesting node
		"""
		return sender

	def rpc_ping(self, sender: Tuple[str, int], node_id: int) -> int:
		"""
		Accept an incoming request from sender and return sender's ID
		to indicate a successful ping

		Parameters
		----------
			sender: Tuple
				Address of sender that initiated ping
			node_id: int
				ID of sender that initated ping

		Returns
		-------
			int:
				ID of requesting node
		"""
		source = Node(node_id, sender[0], sender[1])
		log.debug("%s got ping request from %s", self.source_node, join_addr(sender))
		self.welcome_if_new(source)
		return self.source_node.digest_id

	def rpc_store(self, sender: "Node", node_id: bytes, key: bytes, value: Any) -> bool:
		"""
		Store data from a given sender

		Parameters
		----------
			sender: Node
				Node that is initiating/requesting store
			node_id: bytes
				ID of node that is initiating/requesting store
			key: bytes
				ID of resource to be stored
			value: Any
				Payload to be stored at `key`

		Returns
		-------
			bool:
				Indicator of successful operation
		"""
		source = Node(node_id, sender[0], sender[1])
		self.welcome_if_new(source)
		# pylint: disable=bad-continuation
		log.debug("%s got store request from %s, storing %iB at %s",
					self.source_node, join_addr(sender), len(value), key.hex())
		resource = Node(key, type=NodeType.Resource, value=value)
		self.storage.set(resource)
		return True

	def rpc_find_node(self,
		sender: "Node",
		node_id: int,
		key: int
	) -> List[Tuple[int, str, int]]:
		"""
		Return a list of peers that are closest to a given key (node_id to be found)

		Parameters
		----------
			sender: Node
				The node initiating the request
			node_id: int
				ID of the node initiating the request
			key: int
				ID node who's closes neighbors we want to return

		Returns
		-------
			List[Tuple[int, str, int]]:
				Addresses of closest neighbors in regards to resource `key`
		"""
		source = Node(node_id, sender[0], sender[1])
		log.info("%s finding neighbors of %s in local table", self.source_node, source)
		self.welcome_if_new(source)
		node = Node(key)
		neighbors = self.router.find_neighbors(node, exclude=source)
		return list(map(tuple, neighbors))

	def rpc_find_value(self,
		sender: "Node",
		node_id: int,
		key: int
	) -> Union[List[Tuple[int, str, int]], Dict[str, Any]]:
		"""
		Return the value at a given key. If the key is found, return it
		to the requestor, else execute an rpc_find_node to find neighbors
		of sender that might have key

		Parameters
		----------
			sender: Node
				Node at which key is stored
			node_id: int
				ID of node at which key is stored
			key: int
				ID of resource to be found

		Returns
		-------
			Union[List[Tuple[int, str, int]], Dict[str, Any]]:
				Will be either the given value indexed in a hashmap if the value is
				found, or will recursively attempt to find node at which key is
				stored via calls to `rpc_find_node`
		"""
		source = Node(node_id, sender[0], sender[1])
		self.welcome_if_new(source)
		value = self.storage.get(key, None)
		if value is None:
			return self.rpc_find_node(sender, node_id, key)
		return {"value": value}

	async def call_find_node(self,
		node_to_ask: "Node",
		node_to_find: "Node"
	) -> List[Tuple[int, str, int]]:
		"""
		Dial a given node_to_ask in order to find node_to_find

		Parameters
		----------
			node_to_ask: Node
				Node to ask regarding node_to_find
			node_to_find: Node
				Node that this call is attempting to find

		Returns
		-------
			List[Tuple[int, str, int]]:
				Nodes closes to node_to_find which to continue search
		"""
		address = (node_to_ask.ip, node_to_ask.port)
		result = await self.find_node(address, self.source_node.digest_id, node_to_find.digest_id)
		return self.handle_call_response(result, node_to_ask)

	async def call_find_value(self,
		node_to_ask: "Node",
		node_to_find: "Node"
	) -> Union[List[Tuple[int, str, int]], Dict[str, Any]]:
		"""
		Dial a given node_to_ask in order to find a value on node_to_find

		Parameters
		----------
			node_to_ask: Node
				Node to ask in order to find node_to_find to retrieve a given value
			node_to_find: Node
				Node that this call is attempting to find

		Returns
		-------
			Union[List[Tuple[int, str, int]], Dict[str, Any]]:
				Either the list of nodes close(r) to the key associated with this
				value, or the actual value
		"""
		address = (node_to_ask.ip, node_to_ask.port)
		result = await self.find_value(address, self.source_node.digest_id, node_to_find.id)
		return self.handle_call_response(result, node_to_ask)

	async def call_ping(self, node_to_ask: "Node") -> int:
		"""
		Wrapper for rpc_ping, where we just handle the result

		Parameters
		----------
			node_to_ask: Node
				Node at which to send ping request

		Returns
		-------
			int:
				ID of peer responding to ping
		"""
		address = (node_to_ask.ip, node_to_ask.port)
		result = await self.ping(address, self.source_node.digest_id)
		return self.handle_call_response(result, node_to_ask)

	async def call_store(self, node_to_ask: "Node", key: int, value: Any) -> bool:
		"""
		Wrapper for rpc_store, where we handle the result

		Parameters
		----------
			node_to_ask: Node
				Node which to ask to store a given key/value pair
			key: int
				ID of resource to store
			value: Any
				Payload to store at key address

		Returns
		-------
			bool:
				Indication that store operation was succesful
		"""
		address = (node_to_ask.ip, node_to_ask.port)
		result = await self.store(address, self.source_node.digest_id, key, value)
		return self.handle_call_response(result, node_to_ask)

	def welcome_if_new(self, node: "Node"):
		"""
		Given a new node (Peer), send it all the keys/values it should be storing,
		then add it to the routing table.

		Process:
			For each key in storage, get k closest nodes.  If newnode is closer
			than the furtherst in that list, and the node for this server
			is closer than the closest in that list, then store the key/value
			on the new node (per section 2.5 of the paper)

		Parameters
		----------
			node: Node
				Node to add to routing table
		"""
		if not self.router.is_new_node(node) or node.type == NodeType.Resource:
			return

		log.info("%s welcoming new node %s", self.source_node, node)

		for inode in self.storage:
			neighbors = self.router.find_neighbors(inode)

			if neighbors:
				furthest = neighbors[-1].distance_to(inode)
				is_closer_than_furthest = node.distance_to(inode) < furthest
				closest_distance_to_new = neighbors[0].distance_to(inode)
				curr_distance_to_new = self.source_node.distance_to(inode) < closest_distance_to_new

			if not neighbors or (is_closer_than_furthest and curr_distance_to_new):
				asyncio.ensure_future(self.call_store(node, inode.digest_id, inode.value))

		self.router.add_contact(node)

	def handle_call_response(self, result: Any, node: "Node"):
		"""
		If we get a valid response, welcome the node (if need be). If
		we get no response, remove the node as peer is down

		Parameters
		----------
			result: Any
				Could be the result of any rpc method
			node: Node
				Node to which operation was sent

		Returns
		-------
			result: Any
				The result from our rpc method
		"""
		if not result[0]:
			# pylint: disable=bad-continuation
			log.warning("%s got no response from %s, removing from router",
							self.source_node, node)
			self.router.remove_contact(node)
			return result

		log.info("%s handling successful response from %s", self.source_node, node)
		self.welcome_if_new(node)
		return result



import logging
from collections import Counter
import asyncio
from typing import List, Any, Dict

from liaa.node import NodeHeap
from liaa.protocol import RPCFindResponse
from liaa.utils import gather_dict

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


# pylint: disable=too-few-public-methods
class SpiderCrawl:
	# pylint: disable=bad-continuation
	def __init__(self,
		protocol: "KademliaProtocol",
		node: "Node",
		peers: List["Node"],
		ksize: int,
		alpha: int
	):
		"""
		The C{SpiderCrawl}er is a base class that is responsible for bootstrapping
		various sub-classes (sub-crawlers) with a list of necessary functions,
		including _find and _nodes_found methods

		Parameters
		-----------
			protocol: KademliaProtocol
				A (kademlia) protocol instance.
			node: Node
				representing the key we're looking for
			peers: List[Node]
				A list of instances that provide the entry point for the network
			ksize: int
				The value for k based on the paper
			alpha: int
				The value for alpha based on the paper
		"""
		self.protocol = protocol
		self.ksize = ksize
		self.alpha = alpha
		self.node = node
		self.nearest = NodeHeap(self.node, self.ksize)
		self.last_ids_crawled = []
		self.nearest.push(peers)

		log.info("%s creating spider with %i peers", self.node, len(peers))

	async def _find(self, rpcmethod: asyncio.Future) -> asyncio.Future:
		"""
		Make a either a call_find_value or call_find_node rpc to our nearest
		neighbors in attempt to find some node

		Parameters
		-----------
			rpcmethod: asyncio.Future
				The protocol's call_find_value or call_find_node method

		The process:
		  1. calls find_* to current ALPHA nearest not already queried nodes,
			 adding results to current nearest list of k nodes.
		  2. current nearest list needs to keep track of who has been queried
			 already sort by nearest, keep KSIZE
		  3. if list is same as last time, next call should be to everyone not
			 yet queried
		  4. repeat, unless nearest list has all been queried, then ur done

		Returns
		-------
			asyncio.Future:
				_nodes_found callback, which should be overloaded in sub-classes
		"""
		# pylint: disable=bad-continuation
		log.info("%s making find with %s on nearest: %s", self.node,
				rpcmethod.__name__, ",".join(map(str, self.nearest)))
		count = self.alpha
		if self.nearest.get_ids() == self.last_ids_crawled:
			count = len(self.nearest)
		self.last_ids_crawled = self.nearest.get_ids()

		dicts = {}
		for peer in self.nearest.get_uncontacted()[:count]:
			dicts[peer.digest_id] = rpcmethod(peer, self.node)
			self.nearest.mark_contacted(peer)
		found = await gather_dict(dicts)
		return await self._nodes_found(found)

	async def _nodes_found(self, responses):
		"""
		A callback to execute once nodes are found via _find
		"""
		raise NotImplementedError


class ValueSpiderCrawl(SpiderCrawl):
	# pylint: disable=bad-continuation
	def __init__(self,
		protocol: "KademliaProtocol",
		node: "Node",
		peers: List["Node"],
		ksize: int,
		alpha: int
	):
		"""
		The C{ValueCrawl}er is basically responsible for executing recursive calls
		to our _find method, which searches our nearest nodes (and the nearest nodes
		to those nodes, so on and so forth) in an attempt to find a given 160-bit
		resource key. This crawler will either return a callback to _handle_found_values
		if values for the given key are found, or the crawler will return None
		if the given key cannot be found via our current node

		Parameters
		----------
			protocol: KademliaProtocol
				A (kademlia) protocol instance.
			node: Node
				representing the key we're looking for
			peers: List[Node]
				A list of instances that provide the entry point for the network
			ksize: int
				The value for k based on the paper
			alpha: int
				The value for alpha based on the paper


		"""
		super(ValueSpiderCrawl, self).__init__(protocol, node, peers, ksize, alpha)

		# keep track of the single nearest node without value - per
		# section 2.3 so we can set the key there if found
		self.nearest_without_value = NodeHeap(self.node, 1)

	async def find(self):
		"""
		A wrapper for the base class's _find, where we attempt to find the
		closest value requested using the protocols call_find_value rpc method
		Returns
		-------
			Optional[Any]:
				Where Any is either:
					(1) _find, if we did not find key, but have peers left to search
					(2) None, if we've searched all peers without finding key
					(3) _handle_found_values, if we found values related to our key
		"""
		return await self._find(self.protocol.call_find_value)

	async def _nodes_found(self, responses: Dict[str, Any]):
		"""
		Recursively execute a _find and handle all returned values. These values
		can be nodes representing closer nodes to our eventual destination as well
		as the potential values that we've found related to our key

		Parameters
		----------
			responses: List[Tuple[bool, Union[List[Tuple[int, str, int]], Dict[str, Any]]]]
				Responses from _find

		Returns
		-------
			Optional[asyncio.Future]
				Which can be either:
					(1) a recursive call to _find if we have more searching to do
					(2) None, if we've exhausted our search without finding our key
					(3) A call to _handle_found_values if we've found values
		"""
		to_remove = []
		found_values = []
		for peer_id, response in responses.items():
			response = RPCFindResponse(response)
			if not response.did_happen():

				# if we did not get a response from the peer in question,
				# we need to remove this peer from our nearest as a way of
				# pruning the network (nodes that don't have resources get
				# lower priority)
				to_remove.append(peer_id)
			elif response.has_value():

				# if we found the value handle it accordingly
				found_values.append(response.get_value())
			else:

				# if we got a response but did not find a value, keep note
				# of this peer not having the value we're searching for, and
				# add this peer's nearest peers to the current node's nearest peers
				peer = self.nearest.get_node(peer_id)
				self.nearest_without_value.push(peer)
				self.nearest.push(response.get_node_list())

		# prune our list of nearest nodes
		self.nearest.remove(to_remove)

		# if our search returned values, handle them
		if found_values:
			return await self._handle_found_values(found_values)

		# if we've contacted all nodes in our binary tree (heapq) and
		# have found no value we do nothing (the network has been updated
		# accordingly already)
		if self.nearest.have_contacted_all():
			return None

		# recursively execute find until we return from either _handle_found_values
		# or from have_contacted_all
		return await self.find()

	async def _handle_found_values(self, values: List[str]):
		"""
		We got some values!  Exciting.  But let's make sure they're all the
		same or freak out a little bit.  Also, make sure we tell the nearest
		node that *didn't* have the value to store it.

		Basically this method is responsible for caching found values closer
		to the current node so as to increase the performance/lookup of
		network operations

		Parameters
		----------
			values: List[str]
				Values returned from recursive _find operation

		Returns
		-------
			value: Any
				Original value that we were searching for
		"""
		value_counts = Counter(values)
		if len(value_counts) != 1:
			log.warning("%s multiple values for %s", self.node, str(values))
		value = value_counts.most_common(1)[0][0]

		peer = self.nearest_without_value.popleft()
		if peer:
			log.debug("%s asking nearest node %i to store %s", self.node, peer.long_id, str(value))
			await self.protocol.call_store(peer, self.node.digest_id, value)
		return value


class NodeSpiderCrawl(SpiderCrawl):
	async def find(self):
		"""
		A wrapper for the base class's _find, where we attempt to find the
		closest node requested using the protocols call_find_node rpc method

		Returns
		-------
			Callable[[Callable[None, Any]], Any]:
				Where Any is either:
					(1) A recursive call to _find
					(2) A call to _nodes_found
					(3) None, if we've searched all peers without finding key
		"""
		return await self._find(self.protocol.call_find_node)

	async def _nodes_found(self, responses: List[Any]) -> asyncio.Future:
		"""
		Handle the result of an iteration in _find.

		Parameters
		----------
			responses: Any
				Responses received from peer nodes via NodeCrawl execution

			Where Any can include:
				List[Tuple[int, str, int]] representing peer nodes
				Dict[str, Any] representing found values

		Returns
		-------
			asyncio.Future:
				Recursive call to _find
		"""
		toremove = []
		for peerid, response in responses.items():
			response = RPCFindResponse(response)
			if not response.did_happen():
				log.debug("%s encountered empty response, removing...", self.node)
				toremove.append(peerid)
			else:
				self.nearest.push(response.get_node_list())
		self.nearest.remove(toremove)

		if self.nearest.have_contacted_all():
			log.debug("%s has contacted all nearest nodes", self.node)
			return list(self.nearest)
		return await self.find()



import sys
import operator
import random
import string
import asyncio
from typing import Dict, Any, List, Union, Tuple, Optional


def join_addr(addr: Tuple[str, int]) -> str:
	"""
	Join a tuple address to string

	Parameters
	----------
		addr: Tuple[str, int]
			Address to be joined

	Returns
	-------
		str:
			String of address
	"""
	return ":".join(map(str, addr))


def split_addr(addr: "str") -> Tuple[str, int]:
	"""
	Split a string address to tuple

	Parameters
	----------
		addr: str
			Address to transform

	Returns
	-------
		Tuple[str, int]:
			Tuple version of input address
	"""
	host, port = addr.split(":")
	return host, int(port)


def rand_str(num=20) -> str:
	"""
	Create a random string array
	"""
	chars = string.ascii_letters + string.digits
	return "".join(random.sample(chars, num))


def rand_int_id() -> str:
	"""
	Create a random string array

	Returns
	-------
		int:
			160-bit integer
	"""
	return random.randint(1, 2**160)

def rand_digest_id():
	num = rand_int_id()
	return int_to_digest(num)


def int_to_hex(num: int) -> str:
	return int_to_digest(num).hex()


async def gather_dict(dic: Dict[str, Any]) -> Dict[str, Any]:
	"""
	Execute a list of coroutines and return the results
	in a hashmap

	Parameters
	----------
		dic: Dict[str, Any]
			Hashmap of coroutines to be execute

	Returns
	-------
		Dict[str, Any]:
			Results of executed coroutines
	"""
	cors = list(dic.values())
	results = await asyncio.gather(*cors)
	return dict(zip(dic.keys(), results))


def digest_to_int(byte_arr: bytes) -> int:
	return hex_to_int(byte_arr.hex())


def int_to_digest(num: int) -> bytes:
	return num.to_bytes((num.bit_length() // 8) + 1, byteorder='big')


def shared_prefix(args: List[str]) -> str:
	"""
	Find the shared prefix between the strings.

	Parameters
	----------
		*args:
			Variable length

	Example
	-------
		assert 'blah' == shared_prefix(['blahblah', 'blahwhat'])
	"""
	i = 0
	while i < min(map(len, args)):
		if len(set(map(operator.itemgetter(i), args))) != 1:
			break
		i += 1
	return args[0][:i]


def bytes_to_bit_string(arr: bytes) -> str:
	"""
	Convert a byte array to a bit array

	Parameters
	----------
		arr: bytes
			Bytes to be transformed

	Returns
	-------
		str:
			bit array of input
	"""
	bits = [bin(bite)[2:].rjust(8, '0') for bite in arr]
	return "".join(bits)


def hex_to_int_digest(hexval: str) -> bytes:
	return int_to_digest(hex_to_int(hexval))


def hex_to_int(hexval: bytes, base: int = 16) -> int:
	"""
	Convert given hex to a base-16 integer

	Parameters
	----------
		hexval: bytes
			Hex byte array
		base: int
			Base of integer conversion (default=16)
	"""
	return int(hexval, base)


def check_dht_value_type(value: Any) -> bool:
	"""
	Checks to see if the type of the value is a valid type for
	placing in the dht.

	Parameters
	----------
		value: Any
			Value to be checked

	Returns
	-------
		bool:
			Indicating whether or not type is in DHT types
	"""
	# pylint: disable=bad-continuation
	typeset = [
		int,
		float,
		bool,
		str,
		bytes
	]
	return type(value) in typeset  # pylint: disable=unidiomatic-typecheck


class ArgsParser:
	def __init__(self):
		"""
		Util used to parse command line args on app start


		"""
		self._data: Dict[str, Union[str, int, bool]] = {}

	def add_many(self, opts: List[Tuple[str, str]]) -> None:
		"""
		Add command line args to ArgsParser.data

		Parameters
		-----------
			opts: List[Tuple[str, str]]
				List of tuple key,value pairs to add


		"""
		for key, val in opts:
			self._data[key] = val

	def get(self, key: str, alt_key: str) -> Optional[str]:
		"""
		Get a key from the ArgsParser, or get the alt_key if the key
		isnt present, else log and exit

		Parameters
		----------
			key: str
				First option key to fetch
			alt_key: str
				Second option key to fetch

		Returns
		-------
			Optional[str]:
				Value associated with key/alt_key
		"""
		try:
			return self._data[key]
		except KeyError:
			try:
				return self._data[alt_key]
			except KeyError:
				print(f"opt {key} {alt_key} not in args")
				sys.exit(1)

	def __contains__(self, key: str) -> bool:
		return key in self._data

	# pylint: disable=no-self-use
	def has_help_opt(self):
		return len(self._data) == 1 and "-h" in self._data or "--help" in self._data

	def has_proper_opts(self):
		# pylint: disable=bad-continuation
		required = [
			("-i", "--ip"),
			("-p", "--port"),
		]
		def has_arg(arg):
			return arg[0] in self._data or arg[1] in self._data
		return all(map(has_arg, required)), "missing required opts"


def str_arg_to_bool(arg: str) -> Union[bool, str]:
	"""
	Convert string cli arg to boolean

	Parameters
	----------
		arg: str
			Arg to be converted

	Returns
	-------
		Union[bool, str]:
			bool if arg is false else original arg
	"""
	return False if arg.lower() == 'false' else arg




import datetime as dt
import functools
import logging
import operator
import os
import pickle
import time
from abc import ABC, abstractmethod
from collections import OrderedDict
from collections.abc import Iterable
from itertools import takewhile
from typing import Any, List, Optional, Tuple

from liaa.node import Node, NodeType
from liaa.utils import hex_to_int_digest

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def pre_prune():
	"""
	Decorator (syntactic sugar) for a storage interface's `prune()`
	method
	"""
	def wrapper(func):
		@functools.wraps(func)
		def _pre_prune(*args):
			"""
			Parameters
			----------
				args[0]: IStorage
					Reference to instance of storae interface
			"""
			log.debug("%s pruning items...", args[0].node)
			args[0].prune()
			return func(*args)
		return _pre_prune
	return wrapper


class IStorage(ABC):
	"""
	IStorage

	Local storage for this node.
	IStorage implementations of get must return the same type as put in by set
	"""

	@abstractmethod
	def get(self, hexkey: str, default=None) -> Optional["Node"]:
		pass

	@abstractmethod
	def set(self, node: "Node"):
		pass

	@abstractmethod
	def remove(self, hexkey: str):
		pass

	@abstractmethod
	def iter_older_than(self, seconds_old: int):
		pass

	@abstractmethod
	def prune(self):
		pass

	@abstractmethod
	def __iter__(self):
		pass

	@abstractmethod
	def __contains__(self, hexkey: str):
		pass

	@abstractmethod
	def __len__(self):
		pass


class EphemeralStorage(IStorage):
	def __init__(self, node: "Node", ttl=604800):
		"""
		EphemeralStorage

		Parameters
		----------
			node: Node
				The node representing this peer
			ttl: int
				Max age that items can live untouched before being pruned
				(default=604800 seconds = 1 week)
		"""
		self.node = node
		self.data = OrderedDict()
		self.ttl = ttl

	@pre_prune()
	def get(self, hexkey: str, default: Optional[Any] = None) -> Optional["Node"]:
		"""
		Retrieve a node from storage

		Parameters
		----------
			hexkey: str
				Hex value of node's long_id
			default: Optional[Any]
				Default value to return if node not in storage

		Returns
		-------
			Optional[Node]:
				Node if node is in storage, else `default`
		"""
		log.debug("%s fetching resource %s", self.node, hexkey)
		if hexkey in self:
			_, value = self.data[hexkey]
			return Node(hex_to_int_digest(hexkey), type=NodeType.Resource, value=value)
		log.debug("Resource %s not found on node %s", hexkey, self.node)
		return default

	def set(self, node: "Node"):
		"""
		Save a given Node in storage

		Parameters
		----------
			node: Node
				Node to be saved
		"""
		log.debug("%s setting resource %s", self.node, node.hex)
		self.data[node.hex] = (time.monotonic(), node.value)
		log.debug("%s storage has %i items", self.node, len(self))

	def remove(self, hexkey: str) -> None:
		"""
		Remove a node from storage

		Parameters
		----------
			hexkey: str
				Hex value of node's long_id
		"""
		if hexkey in self:
			log.debug("%s removing resource %s", self.node, hexkey)
			del self.data[hexkey]
			return
		log.debug("Resource %s not found on node %s", hexkey, self.node)

	def prune(self) -> None:
		"""
		Prune storage
		"""
		for _, _ in self.iter_older_than(self.ttl):
			self.data.popitem(last=False)

	def iter_older_than(self, seconds_old: int) -> List[Tuple[int, Any]]:
		"""
		Return nodes that are older than `seconds_old`

		** For EphemeralStorage we use operator.itemgetter(0, 2) in order to
		return just keys and values (without time.monotonic())

		Parameters
		----------
			seconds_old: int
				Time threshold (seconds)

		Returns
		-------
			Iterable:
				Zipped keys, and values of nodes that are older that `seconds_old`
		"""
		min_birthday = time.monotonic() - seconds_old
		zipped = self._triple_iter()
		matches = takewhile(lambda r: min_birthday >= r[1], zipped)
		items = list(map(operator.itemgetter(0, 2), matches))
		log.debug("%s returning %i nodes for republishing....", self.node, len(items))
		return items

	def _triple_iter(self) -> Iterable:
		"""
		Iterate over EphermeralStorage to return each contents key, time,
		and values
		"""
		ikeys = self.data.keys()
		ibirthday = map(operator.itemgetter(0), self.data.values())
		ivalues = map(operator.itemgetter(1), self.data.values())
		return zip(ikeys, ibirthday, ivalues)

	@pre_prune()
	def __repr__(self) -> str:
		return repr(self.data)

	@pre_prune()
	def __iter__(self) -> Iterable:
		log.debug("%s iterating over %i items in storage", self.node, len(self.data))
		ikeys = self.data.keys()
		ivalues = map(operator.itemgetter(1), self.data.values())
		# pylint: disable=bad-continuation
		nodes = [Node(hex_to_int_digest(p[0]),
					type=NodeType.Resource,
					value=p[1])
			for p in zip(ikeys, ivalues)]
		for node in nodes:
			yield node

	def __contains__(self, hexkey: str) -> bool:
		return hexkey in self.data

	@pre_prune()
	def __len__(self) -> int:
		return len(self.data)


class DiskStorage(IStorage):
	# pylint: disable=bad-continuation
	def __init__(self, node: "Node", ttl=604800):
		"""
		DiskStorage

		Parameters
		----------
			node: Node
				The node representing this peer
			ttl: int
				Max age that items can live untouched before being pruned
				(default=604800 seconds = 1 week)
		"""
		self.node = node
		self.ttl = ttl

		kstore_dir = os.path.join(os.path.expanduser("~"), ".liaa")
		if not os.path.exists(kstore_dir):
			log.debug("Liaa dir at %s not found, creating...", kstore_dir)
			os.mkdir(kstore_dir)

		self.dir = os.path.join(kstore_dir, str(self.node.long_id))
		if not os.path.exists(self.dir):
			log.debug("Node dir at %s not found, creating...", self.dir)
			os.mkdir(self.dir)

		self.content_dir = os.path.join(self.dir, "content")
		if not os.path.exists(self.content_dir):
			log.debug("Node content dir at %s not found, creating...", self.content_dir)
			os.mkdir(self.content_dir)

	@pre_prune()
	def get(self, hexkey: str, default: Optional[Any] = None) -> Optional["Node"]:
		"""
		Retrieve a node from storage

		Parameters
		----------
			hexkey: str
				Hex value of node's long_id
			default: Optional[Any]
				Default value to return if node not in storage

		Returns
		-------
			Optional[Node]:
				Node if node is in storage, else `default`
		"""
		log.debug("%s fetching resource %s", self.node, hexkey)
		if hexkey in self:
			# pylint: disable=bad-continuation
			return Node(hex_to_int_digest(hexkey),
						type=NodeType.Resource,
						value=self._load_data(hexkey))
		log.debug("Resource %s not found on node %s", hexkey, self.node)
		return default

	def set(self, node: "Node") -> None:
		"""
		Save a given Node in storage

		Parameters
		----------
			node: Node
				Node to be saved
		"""
		if node.hex in self:
			self.remove(node)
		log.debug("%s setting resource %s", self.node, node.hex)
		self._persist_data(node)
		log.debug("%s storage has %i items", self.node, len(self))

	def remove(self, hexkey: str) -> None:
		"""
		Remove a node from storage

		Parameters
		----------
			hexkey: str
				Hex value of node's long_id
		"""
		try:
			fname = os.path.join(self.content_dir, hexkey)
			log.debug("%s removing resource %s", self.node, hexkey)
			os.remove(fname)
		except FileNotFoundError as err:
			log.debug("Resource %s not found on node %s: %s", hexkey, self.node, str(err))

	def iter_older_than(self, seconds_old: int) -> Iterable:
		"""
		Return nodes that are older than `seconds_old`

		Parameters
		----------
			seconds_old: int
				Time threshold (seconds)

		Returns
		-------
			Iterable:
				Zipped keys, and values of nodes that are older that `seconds_old`
		"""
		to_republish = filter(lambda t: t[1] > seconds_old, self._content_stats())
		repub_keys = list(map(operator.itemgetter(0), to_republish))
		repub_data = [self._load_data(k) for k in repub_keys]
		log.debug("%s returning %i nodes for republishing....", self.node, len(repub_keys))
		return zip(repub_keys, repub_data)

	def prune(self) -> None:
		"""
		Prune storage
		"""
		for key, _ in self.iter_older_than(self.ttl):
			self.remove(key)

	def contents(self) -> List[str]:
		"""
		List all nodes in storage

		TODO: ideally, we shouldn't have to filter out the state file
		like this, we should maybe keep all config/state files in the parent
		directory, and the actual data files in a sub-directory

		Returns
		-------
			List[str]:
				Contents of storage directory
		"""
		# up until -1 will prevent node.state from being loaded
		return os.listdir(self.content_dir)

	def _persist_data(self, node: "Node") -> None:
		"""
		Save a given node's value to disk

		Parameters
		----------
			node: Node
				The node to save
		"""
		fname = os.path.join(self.content_dir, node.hex)
		log.debug("%s attempting to persist %s", self.node, node.hex)
		data = {"value": node.value, "time": time.monotonic()}
		with open(fname, "wb") as ctx:
			pickle.dump(data, ctx)

	def _load_data(self, hexkey: str) -> Optional[Any]:
		"""
		Load a data at a given hexkey

		Parameters
		----------
			hexkey: str
				Hexkey to load

		Returns
		-------
			Optional[Any]:
				Data if hexkey is found, else None
		"""
		fname = os.path.join(self.content_dir, hexkey)
		log.debug("%s attempting to read resource node at %s", self.node, hexkey)
		try:
			with open(fname, "rb") as ctx:
				data = pickle.load(ctx)
				return data["value"]
		except FileNotFoundError as err:
			log.error("%s could not load key at %s: %s", self.node, hexkey, str(err))

	def _content_stats(self) -> List[Tuple[str, float]]:
		"""
		For each node in storage, return its 'last modified time'

		Returns
		-------
			List[Tuple[str, float]]
				List of (filename, last_modified_time) pairs
		"""
		def time_delta(hexkey: str) -> Tuple[str, float]:
			path = os.path.join(self.content_dir, hexkey)
			statbuff = os.stat(path)
			diff = dt.datetime.fromtimestamp(time.time()) - dt.datetime.fromtimestamp(statbuff.st_mtime)
			return hexkey, diff.seconds
		return list(map(time_delta, self.contents()))

	@pre_prune()
	def __iter__(self) -> Iterable:
		log.debug("%s iterating over %i items in storage", self.node, len(self.contents()))
		ikeys = self.contents()
		ivalues = [self._load_data(k) for k in ikeys]
		# pylint: disable=bad-continuation
		nodes = [Node(hex_to_int_digest(p[0]),
					type=NodeType.Resource,
					value=p[1])
					for p in zip(ikeys, ivalues)]
		for node in nodes:
			yield node

	def __contains__(self, hexkey: str) -> bool:
		return hexkey in self.contents()

	@pre_prune()
	def __repr__(self) -> str:
		return repr(self.contents())

	@pre_prune()
	def __len__(self) -> int:
		return len(self.contents())


StorageIface = DiskStorage



import sys
import os

__version__ = "1.1.2"

WORKDIR = os.path.basename(os.path.basename(os.path.abspath(__file__)))
if not WORKDIR in sys.path:
	sys.path.insert(0, WORKDIR)

if sys.version_info.major < 3 or sys.version_info.minor < 7:
	raise RuntimeError("Python version >= 3.7 required")



import heapq
import time
import operator
import asyncio

from itertools import chain
from collections import OrderedDict
from liaa.utils import shared_prefix, bytes_to_bit_string


class KBucket:
	def __init__(self, range_lower, range_upper, ksize):
		self.range = (range_lower, range_upper)
		self.nodes = OrderedDict()
		self.replacement_nodes = OrderedDict()
		self.touch_last_updated()
		self.ksize = ksize

	def touch_last_updated(self):
		self.last_updated = time.monotonic()

	def get_nodes(self):
		return list(self.nodes.values())

	def split(self):
		midpoint = (self.range[0] + self.range[1]) / 2
		one = KBucket(self.range[0], midpoint, self.ksize)
		two = KBucket(midpoint + 1, self.range[1], self.ksize)
		nodes = chain(self.nodes.values(), self.replacement_nodes.values())
		for node in nodes:
			bucket = one if node.long_id <= midpoint else two
			bucket.add_node(node)

		return (one, two)

	def remove_node(self, node):
		if node.digest_id in self.replacement_nodes:
			del self.replacement_nodes[node.digest_id]

		if node.digest_id in self.nodes:
			del self.nodes[node.digest_id]

			if self.replacement_nodes:
				newnode_id, newnode = self.replacement_nodes.popitem()
				self.nodes[newnode_id] = newnode

	def has_in_range(self, node):
		return self.range[0] <= node.long_id <= self.range[1]

	def is_new_node(self, node):
		return node.digest_id not in self.nodes

	def add_node(self, node):
		"""
		Add a C{Node} to the C{KBucket}.  Return True if successful,
		False if the bucket is full.

		If the bucket is full, keep track of node in a replacement list,
		per section 4.1 of the paper.
		"""
		if node.digest_id in self.nodes:
			del self.nodes[node.digest_id]
			self.nodes[node.digest_id] = node
		elif len(self) < self.ksize:
			self.nodes[node.digest_id] = node
		else:
			if node.digest_id in self.replacement_nodes:
				del self.replacement_nodes[node.digest_id]
			self.replacement_nodes[node.digest_id] = node
			return False
		return True

	def depth(self):
		vals = self.nodes.values()
		sprefix = shared_prefix([bytes_to_bit_string(n.digest_id) for n in vals])
		return len(sprefix)

	def head(self):
		return list(self.nodes.values())[0]

	def __getitem__(self, node_id):
		return self.nodes.get(node_id, None)

	def __len__(self):
		return len(self.nodes)


class TableTraverser:
	def __init__(self, table, start_node):
		index = table.get_bucket_index_for(start_node)
		table.buckets[index].touch_last_updated()
		self.current_nodes = table.buckets[index].get_nodes()
		self.left_buckets = table.buckets[:index]
		self.right_buckets = table.buckets[(index + 1):]
		self.left = True

	def __iter__(self):
		return self

	def __next__(self):
		"""
		Pop an item from the left subtree, then right, then left, etc.
		"""
		if self.current_nodes:
			return self.current_nodes.pop()

		if self.left and self.left_buckets:
			self.current_nodes = self.left_buckets.pop().get_nodes()
			self.left = False
			return next(self)

		if self.right_buckets:
			self.current_nodes = self.right_buckets.pop(0).get_nodes()
			self.left = True
			return next(self)

		raise StopIteration


class RoutingTable:
	def __init__(self, protocol, ksize, node):
		"""
		@param node: The node that represents this server.  It won't
		be added to the routing table, but will be needed later to
		determine which buckets to split or not.
		"""
		self.node = node
		self.protocol = protocol
		self.ksize = ksize
		self.flush()

	def flush(self):
		self.buckets = [KBucket(0, 2 ** 160, self.ksize)]

	def split_bucket(self, index):
		one, two = self.buckets[index].split()
		self.buckets[index] = one
		self.buckets.insert(index + 1, two)

	def lonely_buckets(self):
		"""
		Get all of the buckets that haven't been updated in over
		an hour.
		"""
		hrago = time.monotonic() - 3600
		return [b for b in self.buckets if b.last_updated < hrago]

	def remove_contact(self, node):
		index = self.get_bucket_index_for(node)
		self.buckets[index].remove_node(node)

	def is_new_node(self, node):
		index = self.get_bucket_index_for(node)
		return self.buckets[index].is_new_node(node)

	def add_bucket(self, bucket):
		self.buckets.append(bucket)

	def add_contact(self, node):
		index = self.get_bucket_index_for(node)
		bucket = self.buckets[index]

		# this will succeed unless the bucket is full
		if bucket.add_node(node):
			return

		# Per section 4.2 of paper, split if the bucket has the node
		# in its range or if the depth is not congruent to 0 mod 5
		if bucket.has_in_range(self.node) or bucket.depth() % 5:
			self.split_bucket(index)
			self.add_contact(node)
		else:
			asyncio.ensure_future(self.protocol.call_ping(bucket.head()))

	def get_bucket_index_for(self, node):
		"""
		Get the index of the bucket that the given node would fall into.
		"""
		index = None
		for i, bucket in enumerate(self.buckets):
			if node.long_id < bucket.range[1]:
				index = i
				break
		return index

	def get_bucket_for(self, node):
		idx = self.get_bucket_index_for(node)
		return self.buckets[idx]

	def find_neighbors(self, node, k=None, exclude=None):
		k = k or self.ksize
		nodes = []
		for neighbor in TableTraverser(self, node):
			notexcluded = exclude is None or not neighbor.is_same_node(exclude)
			if neighbor.digest_id != node.digest_id and notexcluded:
				heapq.heappush(nodes, (node.distance_to(neighbor), neighbor))
			if len(nodes) == k:
				break

		return list(map(operator.itemgetter(1), heapq.nsmallest(k, nodes)))



#!/usr/bin/env python
from setuptools import setup, find_packages
import liaa

# pylint: disable=bad-continuation
setup(
		name="liaa",
		version=liaa.__version__,
		# pylint: disable=line-too-long
		description="Liaa is a distributed hash table for decentralized peer-to-peer computer networks.",
		long_description=open("README.md", encoding='utf-8').read(),
		long_description_content_type='text/markdown',
		author="Brian Muller, Rashad Alston",
		author_email="bamuller@gmail.com, rashadalston@gmail.com",
		license="MIT",
		url="http://github.com/ralston3/liaa",
		packages=find_packages(),
		install_requires=open("requirements.txt").readlines(),
		classifiers=[
			"Development Status :: 5 - Production/Stable",
			"Intended Audience :: Developers",
			"License :: OSI Approved :: MIT License",
			"Operating System :: OS Independent",
			"Programming Language :: Python",
			"Programming Language :: Python :: 3",
			"Programming Language :: Python :: 3.5",
			"Programming Language :: Python :: 3.6",
			"Programming Language :: Python :: 3.7",
			"Topic :: Software Development :: Libraries :: Python Modules",
		],
)



import logging
import asyncio
import sys
import getopt

from liaa.network import Server
from liaa.utils import ArgsParser, split_addr, str_arg_to_bool


def usage():
	return """
Usage: python app.py -p [port] -n [bootstrap neighbors]
-p --port
	Port on which to listen (e.g., 8000)
-n --neighbors
	Neighbors with which to bootstrap (e.g., 177.91.19.1:8000,178.31.13.21:9876)
		or 'False' if not passing an args
	"""

def main():

	handler = logging.StreamHandler()
	formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
	handler.setFormatter(formatter)
	log = logging.getLogger('liaa')
	log.addHandler(handler)
	log.setLevel(logging.DEBUG)

	loop = asyncio.get_event_loop()
	loop.set_debug(True)

	parser = ArgsParser()

	try:
		opts, _ = getopt.getopt(sys.argv[1:], "p:n:", ["--port", "--neighbors"])
		parser.add_many(opts)
	except getopt.GetoptError as err:
		log.error("GetoptError: %s", err)
		print(usage())
		sys.exit(1)

	if parser.has_help_opt() or not parser.has_proper_opts():
		print(usage())
		sys.exit(1)

	server = Server()
	loop.run_until_complete(server.listen(int(parser.get("-p", "--port"))))

	bootstrap_peers = str_arg_to_bool(parser.get("-n", "--neighbors"))
	if bootstrap_peers:
		bootstrap_peers = bootstrap_peers.split(",")
		loop.run_until_complete(server.bootstrap(list(map(split_addr, bootstrap_peers))))

	try:
		loop.run_forever()
	except KeyboardInterrupt:
		print("\nAttempting to gracefully shut down...")
	finally:
		server.stop()
		print("Shutdown successul")


if __name__ == "__main__":

	main()



from glob import glob

import pycodestyle
import pytest
from pylint import epylint as lint


class LintError(Exception):
	pass


class TestCodeLinting:
	# pylint: disable=no-self-use
	@pytest.mark.skip(reason="test linting on a as-needed basis")
	def test_pylint(self):
		(stdout, _) = lint.py_run('liaa', return_std=True)
		errors = stdout.read()
		if errors.strip():
			raise LintError(errors)

	# pylint: disable=no-self-use
	@pytest.mark.skip(reason="test linting on a as-needed basis")
	def test_pep8(self):
		style = pycodestyle.StyleGuide()
		files = glob('liaa/**/*.py', recursive=True)
		result = style.check_files(files)
		if result.total_errors > 0:
			raise LintError("Code style errors found.")



import asyncio
import os

from liaa.network import Server
from liaa.protocol import KademliaProtocol
# from liaa.node import Resource
from liaa.node import Node, NodeType
from liaa.utils import rand_str, rand_digest_id


PORT = 8765


class TestServer:
	# pylint: disable=no-self-use
	def test_server_instance_is_ok(self):
		server = Server()
		assert isinstance(server, Server)

	def test_server_can_start_and_stop(self):
		loop = asyncio.get_event_loop()
		server = Server()

		assert not server.udp_transport
		assert not server.protocol
		assert not server.refresh_loop
		assert not server.save_state_loop
		assert not server.listener

		loop.run_until_complete(server.listen(PORT))

		assert server.udp_transport
		assert server.protocol
		assert isinstance(server.refresh_loop, asyncio.Handle)
		assert isinstance(server.protocol, KademliaProtocol)
		assert isinstance(server.listener, asyncio.AbstractServer)

		server.stop()

		# assert server.udp_transport.closed()
		assert server.refresh_loop.cancelled()
		assert server.save_state_loop.cancelled()
		# assert not server.listener.is_serving()

	def test_create_protocol_is_interchangeable(self):
		server = Server()
		# pylint: disable=protected-access
		proto = server._create_protocol()
		assert isinstance(proto, KademliaProtocol)

		class CoconutProtocol(KademliaProtocol):
			pass

		class HuskServer(Server):
			protocol_class = CoconutProtocol
		husk_server = HuskServer()
		assert isinstance(husk_server._create_protocol(), CoconutProtocol)


	def test_set_digest_returns_void_when_node_has_no_neighbors(self):
		server = Server()
		resource = Node(rand_digest_id(), type=NodeType.Resource, value=rand_str())
		# pylint: disable=protected-access
		server.protocol = server._create_protocol()
		result = asyncio.run(server.set_digest(resource))
		assert not result

	def test_save_state_saves(self, sandbox, mknode):
		server = Server()

		# pylint: disable=unused-argument,bad-continuation
		def bootstrappable_neighbors_stub():
			return [
				mknode(digest_id=rand_digest_id(), ip="0.0.0.0", port=1234),
				mknode(digest_id=rand_digest_id(), ip="0.0.0.0", port=4321)
			]

		box = sandbox(server)
		box.stub("bootstrappable_neighbors", bootstrappable_neighbors_stub)

		server.save_state()

		expected_path = os.path.join(server.storage.dir, "node.state")
		assert os.path.exists(expected_path)
		assert os.path.isfile(expected_path)

		box.restore()

	def test_can_load_state(self, sandbox, mknode):
		server = Server()
		asyncio.set_event_loop(asyncio.new_event_loop())
		# pylint: disable=unused-argument,bad-continuation
		def bootstrappable_neighbors_stub():
			return [
				# make some fake peers
				mknode(digest_id=rand_digest_id(), ip="0.0.0.0", port=1234),
				mknode(digest_id=rand_digest_id(), ip="0.0.0.0", port=4321)
			]

		def bootstrap_stub(addrs):
			return addrs

		box = sandbox(server)
		box.stub("bootstrappable_neighbors", bootstrappable_neighbors_stub)
		box.stub("bootstrap", bootstrap_stub)

		server.save_state()
		loaded_server = server.load_state()
		assert isinstance(loaded_server, Server)

		box.restore()



import random
import time

from liaa.protocol import KademliaProtocol
from liaa.storage import EphemeralStorage


class TestKademliaProtocol:
	# pylint: disable=no-self-use
	def test_can_init_protocol(self, mknode):
		node = mknode(intid=1)
		proto = KademliaProtocol(node, storage=EphemeralStorage(node), ksize=20)
		assert isinstance(proto, KademliaProtocol)

	def test_can_refresh_ids(self, mknode, mkbucket, fake_proto):
		proto = fake_proto()
		for _ in range(5):
			bucket = mkbucket(ksize=proto.router.ksize)
			for _ in range(5):
				node = mknode()
				bucket.add_node(node)
			proto.router.add_bucket(bucket)

		# randomly pick some buckets to make stale
		sample = random.sample(proto.router.buckets, 3)
		for bucket in sample:
			bucket.last_updated = time.monotonic() - 3600

		to_refresh = proto.get_refresh_ids()
		assert isinstance(to_refresh, list)
		assert len(to_refresh) == 3

	def test_rpc_stun_returns_node(self, mknode, fake_proto):
		proto = fake_proto()
		sender = mknode()
		assert sender == proto.rpc_stun(sender)

	def test_rpc_ping_returns_requestors_id(self, mknode, fake_proto, sandbox):
		sender = mknode()
		proto = fake_proto()

		# pylint: disable=unused-argument
		def ping_stub(sender, node_id):
			return sender.digest_id

		# pylint: disable=unused-argument
		def call_store_stub(node_to_ask, key, value):
			return True

		box = sandbox(proto)
		box.stub("call_store", call_store_stub)
		box.stub("rpc_ping", ping_stub)

		source_id = proto.rpc_ping(sender, sender.digest_id)
		assert source_id == sender.digest_id

		box.restore()
